/*
   Copyright (C) 1999 T. Scott Dattalo

This file is part of gpsim.

gpsim is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

gpsim is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with gpsim; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */


#include <string>

#include "command.h"
#include "cmd_echo.h"
#include "misc.h"

// See cmd_echo.h for why this file is basically a skeleton

cmd_echo echo;

static cmd_options cmd_echo_options[] =
{
  {nullptr, 0, 0}
};


cmd_echo::cmd_echo()
  : command("echo", nullptr)
{ 
  brief_doc = "echo \"text\"";

  long_doc = "echo \"text\" - useful for command files\n";

  token_value = 0;
  op = cmd_echo_options; 
}
